/*
  portfwd.c

  $Id: portfwd.c,v 1.5 2000/03/07 23:33:27 everton Exp $
 */


#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <syslog.h>
#include "portfwd.h"
#include "util.h"
#include "vector.hpp"
#include "iterator.hpp"
#include "entry.hpp"


extern FILE           *yyin;
extern int            yyparse();
extern int            conf_syntax_errors;
extern vector<entry*> *entry_vector;

const int BUF_SZ = 8192;

void usage(FILE *out) 
{
  const char *prog = get_prog_name();
  fprintf(out, "Usage: %s [<options>]\n"
	  "       -v               | --verbose\n"
	  "       -c <config-file> | --config <config-file>\n",
	  prog);
}

void parse_cmdline(int argc, const char *argv[], const char **config)
{
  const char * const DEFAULT_CONFIG = "portfwd.cfg";
  int opt;
  int option_index = 0;
  struct option long_options[] =
  {
    {"help", 0, 0, 'h'},
    {"config", 1, 0, 'c'},
    {"verbose", 0, 0, 'v'},
    {0, 0, 0, 0}
  };

  *config  = 0;

  for (;;) {
		 
    opt = getopt_long(argc, (char ** const) argv, "hc:v", long_options, &option_index);
    if (opt == -1)
      break;

    switch (opt) {
    case 'h':
      usage(stdout);
      exit(0);
      break;
    case 'c':
      if (*config) {
	syslog(LOG_ERR, "Configuration file redefinition: %s\n", optarg);
	exit(1);
      }
      *config = optarg;
      break;
    case 'v':
      ++verbose_mode;
      break;
    case '?':
      usage(stderr);
      exit(1);
      break;
    default:
      syslog(LOG_ERR, "getopt() returned unknown option: %c (%d)\n", opt, opt);
      exit(1);
    }
  }

  if (optind < argc) {
    syslog(LOG_ERR, "Invalid argument: %s\n", argv[optind]);
    exit(1);
  }

  if (!*config)
    *config = DEFAULT_CONFIG;
}

void read_config(const char *cfg)
{
  ONVERBOSE(syslog(LOG_INFO, "Configuration file: '%s'\n", cfg));

  yyin = fopen(cfg, "r");
  if (!yyin) {
    syslog(LOG_ERR, "Can't open configuration file: '%s'\n", cfg);
    exit(1);
  }

  yyparse();

  if (conf_syntax_errors) {
    syslog(LOG_ERR, "Syntax errors: %d", conf_syntax_errors);
    exit(1);
  }
}

void do_show(vector<entry*>* entries)
{
  if (!entries) {
    syslog(LOG_ERR, "Internal error: Null entry vector (!)");
    return;
  }

  if (!entries->get_size()) {
    syslog(LOG_INFO, "/* Empty */");
    return;
  }

  iterator<vector<entry*>,entry*> it(*entries);
  for (it.start(); it.cont(); it.next())
    it.get()->show();
}


void do_forward(vector<entry*>* entries)
{
  if (!entries) {
    syslog(LOG_ERR, "Internal error: Null entry vector (!)");
    return;
  }

  if (!entries->get_size()) {
    syslog(LOG_NOTICE, "Nothing to do: empty configuration file");
    return;
  }

  iterator<vector<entry*>,entry*> it(*entries);
  for (it.start(); it.cont(); it.next())
    it.get()->serve();
}

void child_reaper(int sig)
{
  int status;
  void (*prev_handler)(int);

  /* pid_t child_pid = waitpid(-1, &status, WNOHANG); */
  pid_t child_pid = wait(&status);
  prev_handler = signal(SIGCHLD, child_reaper);
  if (prev_handler == SIG_ERR) {
    syslog(LOG_ERR, "signal() failed on child_reaper re-install: %m");
    return;
  }

  if (child_pid == -1) {
    syslog(LOG_ERR, "Error waiting pid of child: %s\n", strerror(errno));
    return;
  }
  
  if (WIFEXITED(status)) {
    syslog(LOG_WARNING, "Child with PID %d exited normally with exit status: %d\n", child_pid, WEXITSTATUS(status));
    return;
  }

  if (WIFSIGNALED(status))
    syslog(LOG_NOTICE, "Child received signal: %d\n", WTERMSIG(status)); 
    
  syslog(LOG_ERR, "Child with PID %d exited abnormally\n", child_pid);
}

void term_handler(int sig)
{
  void (*prev_handler)(int);

  prev_handler = signal(SIGTERM, SIG_IGN);
  if (prev_handler == SIG_ERR)
    syslog(LOG_ERR, "signal() failed on term_handler de-install: %m");

  if (kill(0, SIGTERM))
    syslog(LOG_ERR, "Can't terminate children\n");

  syslog(LOG_INFO, "SIGTERM - Program terminated\n");

  closelog();

  exit(0);
}

int main(int argc, const char *argv[]) 
{
  const char *config;

  set_prog_name(basename(argv[0]));

  openlog(get_prog_name(), LOG_CONS | LOG_PID, LOG_DAEMON);

  parse_cmdline(argc, argv, &config);

  ONVERBOSE(syslog(LOG_DEBUG, "Verbose mode: %d\n", verbose_mode));

  solve_protonumbers();

  read_config(config);

  ONVERBOSE(do_show(entry_vector));

  if (!verbose_mode) {
    int dmz = daemonize();
    if (dmz) {
      syslog(LOG_ERR, "daemonize() failed: %d\n", dmz);
      exit(1);
    }
  }

  void (*prev_handler)(int);

  prev_handler = signal(SIGCHLD, child_reaper);
  if (prev_handler == SIG_ERR) {
    syslog(LOG_ERR, "signal() failed on child_reaper install: %m");
    exit(1);
  }

  do_forward(entry_vector);

  prev_handler = signal(SIGTERM, term_handler);
  if (prev_handler == SIG_ERR) {
    syslog(LOG_ERR, "signal() failed on term_handler install: %m");
    exit(1);
  }

  for (;;) /* forever */
    sleep(3600);

  return 0;
}
